var prefs = {
  //for messenger toolbar
  icon_mode: 1, //0: icons and text, 1: icons, 2: text, -1: set individual
  placement: [],
  apply_multi: false,
  icon_margin: 0, //margin of icon image and label. TB default is 0.
  //for message window toolbar
  icon_mode_mw: 1,
  placement_mw: [],
  icon_margin_mw: 0,
  //for quick filter toolbar in messenger window
  icon_mode_qf: 0,
  placement_qf: [],
  icon_margin_qf: 0,
  show_cal_cal: true,
  show_todo_cal: true,
  show_today_cal: true
};

//normal: messenger, messageDisplay: messageWindow, normal-QF: quick filter bar in messenger, normal-CAL: calendar buttons in messenger
var currentWindowType = "normal";
var popupId = -1;

browser.windows.onCreated.addListener(async (win) => {
  if (win.type == "normal") {
    //messenger
    await browser.msgHdrCustomizeApi.reorderButtons(win.id, prefs.placement, (prefs.icon_mode == -1));
    if (prefs.icon_mode != -1) await browser.msgHdrCustomizeApi.setIconMode(win.id, prefs.icon_mode);
    await browser.msgHdrCustomizeApi.setIconMargin(win.id, prefs.icon_margin);
    //quick filter
    await browser.msgHdrCustomizeApi.reorderButtonsQuickFilter(prefs.placement_qf, (prefs.icon_mode_qf == -1));
    if (prefs.icon_mode_qf != -1) await browser.msgHdrCustomizeApi.setIconModeQuickFilter(prefs.icon_mode_qf);
    await browser.msgHdrCustomizeApi.setIconMarginQuickFilter(prefs.icon_margin_qf);
    //calendar
    await browser.msgHdrCustomizeApi.showCalenderButtons(prefs["show_cal_cal"], prefs["show_todo_cal"], prefs["show_today_cal"]);
  } else if (win.type == "messageDisplay") {
    //message window
    await browser.msgHdrCustomizeApi.reorderButtons(win.id, prefs.placement_mw, (prefs.icon_mode_mw == -1));
    if (prefs.icon_mode_mw != -1) await browser.msgHdrCustomizeApi.setIconMode(win.id, prefs.icon_mode_mw);
    await browser.msgHdrCustomizeApi.setIconMargin(win.id, prefs.icon_margin_mw);
  }
});

browser.storage.onChanged.addListener(async (changes, area) => {
  let msgrChanged = false;
  let msgwinChanged = false;
  let qfChanged = false;
  let calChanged = false;

  for (let item in changes) {
    prefs[item] = changes[item].newValue;
    msgrChanged = msgrChanged || (item.indexOf("_mw") == -1 && item.indexOf("_qf") == -1 && item.indexOf("_cal") == -1);
    msgwinChanged = msgwinChanged || (item.indexOf("_mw") != -1);
    qfChanged = qfChanged || (item.indexOf("_qf") != -1);
    calChanged = calChanged || (item.indexOf("_cal") != -1);
  }

  if (msgrChanged) {
    await changeHeaderForWindows("normal");
  }

  if (msgwinChanged) {
    await changeHeaderForWindows("messageDisplay");
  }

  if (qfChanged) {
    await changeHeaderForWindows("normal-QF");
  }

  if (calChanged) {
    await changeHeaderForWindows("normal-CAL");
  }

  if (msgrChanged || msgwinChanged || qfChanged) {
    let message = await notifyIconChangeToCustomizeWindow();
    browser.runtime.sendMessage(message);
  }

  return true;
});

async function notifyIconChangeToCustomizeWindow() {
  let buttons = [];
  let popupFound = false;
  let ret = {};

  let wins = await browser.windows.getAll();
  for (let i = 0; i < wins.length; i++) {
    let win = wins[i];
    let winType = currentWindowType.split("-")[0];
    let subType = currentWindowType.split("-")[1];

    if (buttons.length == 0) {
      if (win.type == "normal" && subType == "QF") {
        buttons = await browser.msgHdrCustomizeApi.getButtonsQuickFilter();
      } else if (win.type == "normal" && subType == "CAL") {
        //no need to notify
      } else if (win.type == winType) {
        buttons = await browser.msgHdrCustomizeApi.getButtons(win.id);
      }
    }
    if (win.type == "popup") {
      popupFound = true;
    }
  }

  if (popupFound && buttons.length > 0) {
    ret = {
      message: "BUTTONS_UPDATED",
      buttons: buttons,
      type: currentWindowType
    };
  }

  return ret;
}

browser.runtime.onMessage.addListener((message) => {
  let aPromise = null;
  switch (message.message) {
    case "GET_BUTTONS":
      aPromise = notifyIconChangeToCustomizeWindow().then(message => {
        return message;
      });
      break;
    default:
      aPromise = Promise.resolve(true);
      break;
  }

  return aPromise;
});

async function changeHeaderForWindows(type) {
  let wins = await browser.windows.getAll();
  let subType = type.split("-")[1];
  for (let i = 0; i < wins.length; i++) {
    let win = wins[i];
    if (win.type == "normal" && subType == "QF") {
      await browser.msgHdrCustomizeApi.reorderButtonsQuickFilter(prefs["placement_qf"], (prefs["icon_mode_qf"] == -1));
      if (prefs["icon_mode_qf"] != -1) await browser.msgHdrCustomizeApi.setIconModeQuickFilter(prefs["icon_mode_qf"]);
      await browser.msgHdrCustomizeApi.setIconMarginQuickFilter(prefs["icon_margin_qf"]);
    } else if (win.type == "normal" && subType == "CAL") {
      await browser.msgHdrCustomizeApi.showCalenderButtons(prefs["show_cal_cal"], prefs["show_todo_cal"], prefs["show_today_cal"]);
    } else if (win.type == type) {
      let suffix = (win.type == "messageDisplay") ? "_mw" : "";
      await browser.msgHdrCustomizeApi.reorderButtons(win.id, prefs["placement" + suffix], (prefs["icon_mode" + suffix] == -1));
      if (prefs["icon_mode" + suffix] != -1) await browser.msgHdrCustomizeApi.setIconMode(win.id, prefs["icon_mode" + suffix]);
      await browser.msgHdrCustomizeApi.setIconMargin(win.id, prefs["icon_margin" + suffix]);
    }
  }
}

async function init() {
  for (let key in prefs) {
    let result = await browser.storage.local.get(key);
    let value = result[key];
    if (value !== undefined) {
      prefs[key] = value;
    }
  }

  migrate20xTo21x();

  await browser.msgHdrCustomizeApi.init();

  browser.msgHdrCustomizeApi.onMenuClick.addListener((id) => {
    let url = "html/options_window.html";
    let height = 560;
    if (id.indexOf("MSG") != -1) {
      currentWindowType = "normal";
    } else if (id.indexOf("MW") != -1) {
      currentWindowType = "messageDisplay";
    } else if (id.indexOf("QF") != -1) {
      currentWindowType = "normal-QF";
    } else if (id.indexOf("CAL") != -1) {
      currentWindowType = "normal-CAL";
      url = "html/options_cal.html";
      height = 230;
    }

    browser.windows.get(popupId).then((win) => {
      console.log("Customize window is already opened.");
    }, (reason) => {
      browser.windows.create({
        height: height,
        width: 385,
        url: url,
        type: "popup",
        allowScriptsToClose: true
      }).then((win)=>{
        popupId = win.id;        
      }, (reason)=>{
        console.log("Cannot open customize window.");
        console.log(reason);
      });
    });

  });

  browser.msgHdrCustomizeApi.onLoadMultiselect.addListener(async (id) => {
    if (prefs.apply_multi) {
      await browser.msgHdrCustomizeApi.reorderButtonsMulti(prefs.placement, (prefs.icon_mode == -1));
      if (prefs.icon_mode != -1) await browser.msgHdrCustomizeApi.setIconModeMulti(prefs.icon_mode);
      await browser.msgHdrCustomizeApi.setIconMarginMulti(prefs.icon_margin);
    }
  });

  //for just after install
  changeHeaderForWindows("normal");
  changeHeaderForWindows("messageDisplay");
  changeHeaderForWindows("normal-QF");
  changeHeaderForWindows("normal-CAL");
}

function migrate20xTo21x() {
  for (let i = 0; i < prefs.placement.length; i++) {
    let obj = prefs.placement[i];
    if (obj.iconMode === undefined) {
      obj.iconMode = prefs.icon_mode;
    }

    if (obj.subLabels === undefined) {
      obj.subLabels = [];
    }

    if (obj.renamed === undefined) {
      obj.renamed = false;
    }
  }

  for (let i = 0; i < prefs.placement_mw.length; i++) {
    let obj = prefs.placement_mw[i];
    if (obj.iconMode === undefined) {
      obj.iconMode = prefs.icon_mode_mw;
    }

    if (obj.subLabels === undefined) {
      obj.subLabels = [];
    }

    if (obj.renamed === undefined) {
      obj.renamed = false;
    }
  }

  for (let i = 0; i < prefs.placement_qf.length; i++) {
    let obj = prefs.placement_qf[i];
    if (obj.iconMode === undefined) {
      obj.iconMode = prefs.icon_mode_qf;
    }

    if (obj.subLabels === undefined) {
      obj.subLabels = [];
    }

    if (obj.renamed === undefined) {
      obj.renamed = false;
    }
  }
}

init();

