var {
  ExtensionCommon
} = ChromeUtils.import("resource://gre/modules/ExtensionCommon.jsm");
var {
  Services
} = ChromeUtils.import("resource://gre/modules/Services.jsm");

var {
  ExtensionSupport
} = ChromeUtils.import("resource:///modules/ExtensionSupport.jsm");

var localeStrings = {
  mhcCustomizeToolbar: "",
  mhcForwardInline: "",
  mhcForwardAttach: "",
  mhcForwardTooltip: "",
  mhcForwardInlineTooltip: "",
  mhcForwardAttachTooltip: "",
  mhcForwardAs: "",
  mhcTag: "",
  mhcTagTooltip: "",
  mhcNewTag: "",
  mhcManageTags: "",
  mhcConversation: "",
  mhcConversationTooltip: "",
  mhcCustomizeCalendar: "",
  mhcPrint: "",
  mhcPrintPreview: "",
  mhcPrintTooltip: "",
  mhcPrintPreviewTooltip: "",
  mhcStar: "",
  mhcStarTooltip: ""
};

var defaultOrderMSG = []; //default button order of messenger
var defaultOrderMW = []; //default button order of message window
var defaultOrderQF = []; //default button order of quick filter

//Define API
var msgHdrCustomizeApi = class extends ExtensionCommon.ExtensionAPI {
  getAPI(context) {
    context.callOnClose(this);
    return {
      msgHdrCustomizeApi: {
        init: async function () {
          for (let key in localeStrings) {
            localeStrings[key] = context.extension.localeData.localizeMessage(key);
          }

          return;
        },

        //mode -1: set individually, 0: both, mode 1: icon only, mode 2: text only
        setIconMode: async function (windowId, mode) {
          let rw = null;
          if (windowId != -1) {
            let windowObject = context.extension.windowManager.get(windowId);
            rw = windowObject.window;
          } else {
            rw = Services.wm.getMostRecentWindow("mail:3pane");
          }

          if (rw) {
            let toolbar = rw.document.getElementById("header-view-toolbar");
            if (toolbar) {
              setIconMode(toolbar, mode);
            }
          }
        },

        setIconModeMulti: async function (mode) {
          let rw = Services.wm.getMostRecentWindow("mail:3pane");

          if (rw) {
            //multiselection
            let multimsg = rw.document.getElementById("multimessage");
            let toolbar = multimsg ? multimsg.contentDocument.getElementById("header-view-toolbar") : null;
            if (toolbar) setIconMode(toolbar, mode);
          }
        },

        setIconModeQuickFilter: async function (mode) {
          let rw = Services.wm.getMostRecentWindow("mail:3pane");

          if (rw) {
            //multiselection
            let toolbar = rw.document.getElementById("quick-filter-bar-collapsible-buttons");
            if (toolbar) {
              setIconMode(toolbar, mode);
            }
          }
        },

        setIconMargin: async function (windowId, margin) {
          let rw = null;
          if (windowId != -1) {
            let windowObject = context.extension.windowManager.get(windowId);
            rw = windowObject.window;
          } else {
            rw = Services.wm.getMostRecentWindow("mail:3pane");
          }

          if (rw) {
            let toolbar = rw.document.getElementById("header-view-toolbar");
            if (toolbar) {
              setIconMargin(toolbar, margin);
            }
          }
        },

        setIconMarginMulti: async function (margin) {
          let rw = Services.wm.getMostRecentWindow("mail:3pane");

          if (rw) {
            //multiselection
            let multimsg = rw.document.getElementById("multimessage");
            let toolbar = multimsg ? multimsg.contentDocument.getElementById("header-view-toolbar") : null;
            if (toolbar) setIconMargin(toolbar, margin);
          }
        },

        setIconMarginQuickFilter: async function (margin) {
          let rw = Services.wm.getMostRecentWindow("mail:3pane");

          if (rw) {
            //multiselection
            let toolbar = rw.document.getElementById("quick-filter-bar-collapsible-buttons");
            if (toolbar) {
              setIconMargin(toolbar, margin);
            }
          }
        },
        
        getButtons: async function (windowId) {
          let rw = null;
          if (windowId != -1) {
            let windowObject = context.extension.windowManager.get(windowId);
            rw = windowObject.window;
          } else {
            rw = Services.wm.getMostRecentWindow("mail:3pane");
          }

          let ret = [];
          if (rw) {
            let toolbar = rw.document.getElementById("header-view-toolbar");
            let menupopup = rw.document.getElementById("header-toolbar-context-menu");
            if (menupopup) toolbar.removeChild(menupopup); //For TB89
            ret = getButtonsInToolbar(rw, toolbar);
            if (menupopup) toolbar.insertBefore(menupopup, toolbar.firstChild); //For TB89

            //store original order
            if (rw.location.href.indexOf("messenger.xhtml") != -1) {
              updateDefaultOrder(toolbar, defaultOrderMSG, "tmpMSGIdForRstr");
            } else if (rw.location.href.indexOf("messageWindow.xhtml") != -1) {
              updateDefaultOrder(toolbar, defaultOrderMW, "tmpMWIdForRstr");
            }
          }

          return ret;
        },

        getButtonsQuickFilter: async function () {
          let rw = Services.wm.getMostRecentWindow("mail:3pane");
          let ret = [];
          if (rw) {
            let toolbar = rw.document.getElementById("quick-filter-bar-collapsible-buttons");
            if (toolbar) {
              //store original order. record including separator.
              updateDefaultOrder(toolbar, defaultOrderQF, "tmpQFIdForRstr");
              let separator = toolbar.firstChild;
              toolbar.removeChild(separator);
              ret = getButtonsInToolbar(rw, toolbar);
              toolbar.insertBefore(separator, toolbar.firstChild);
            }
          }

          return ret;
        },

        reorderButtons: async function (windowId, order, setIcon) {
          let rw = null;
          if (windowId != -1) {
            let windowObject = context.extension.windowManager.get(windowId);
            rw = windowObject.window;
          } else {
            rw = Services.wm.getMostRecentWindow("mail:3pane");
          }

          if (rw) {
            let toolbar = rw.document.getElementById("header-view-toolbar");
            if (!toolbar) return false; //rw is other than messenger and message window

            //store original order
            if (rw.location.href.indexOf("messenger.xhtml") != -1) {
              updateDefaultOrder(toolbar, defaultOrderMSG, "tmpMSGIdForRstr");
            } else if (rw.location.href.indexOf("messageWindow.xhtml") != -1) {
              updateDefaultOrder(toolbar, defaultOrderMW, "tmpMWIdForRstr");
            }

            //add custom buttons
            let fwd = rw.document.getElementById("mhcFwdAs");
            if (!fwd) {
              createForwardAsButton(rw, toolbar);
            }
            let tag = rw.document.getElementById("mhcTag");
            if (!tag) {
              createTagButton(rw, toolbar);
            }

            let print = rw.document.getElementById("mhcPrint");
            if (!print) {
              createPrintButton(rw, toolbar);
            }

            let star = rw.document.getElementById("mhcStar");
            if (!star) {
              createStarButton(rw, toolbar);
            }

            /*
            let prev = rw.document.getElementById("mhcPrevMsg");
            if (!prev) {
              createPrevMsgButton(rw, toolbar);
            }
            */
            
            //Add only for messenger window
            if (rw.location.href.indexOf("messenger.xhtml") != -1) {
              let conv = rw.document.getElementById("mhcConv");
              if (!conv) {
                createConversationButton(rw, toolbar);
              }
            }


            if (order.length == 0) return false;

            let menupopup = rw.document.getElementById("header-toolbar-context-menu");
            if (menupopup) toolbar.removeChild(menupopup); //For TB89
            reorderButtonsMain(rw, order, toolbar, setIcon, true);
            if (menupopup) toolbar.insertBefore(menupopup, toolbar.firstChild); //For TB89

            rw.UpdateJunkButton();
            rw.UpdateReplyButtons();
          }

          return true;
        },

        reorderButtonsMulti: async function (order, setIcon) {
          if (order.length == 0) return false;

          let rw = Services.wm.getMostRecentWindow("mail:3pane");

          if (rw) {
            let multimsg = rw.document.getElementById("multimessage");
            let toolbar = multimsg ? multimsg.contentDocument.getElementById("header-view-toolbar") : null;
            if (!toolbar) return false;

            reorderButtonsMain(rw, order, toolbar, setIcon, false);
          }

          return true;
        },

        reorderButtonsQuickFilter: async function (order, setIcon) {
          if (order.length == 0) return false;

          let rw = Services.wm.getMostRecentWindow("mail:3pane");

          if (rw) {
            let toolbar = rw.document.getElementById("quick-filter-bar-collapsible-buttons");
            if (!toolbar) return false;

            //store original order
            updateDefaultOrder(toolbar, defaultOrderQF, "tmpQFIdForRstr");

            let separator = toolbar.firstChild;
            toolbar.removeChild(separator);
            reorderButtonsMain(rw, order, toolbar, setIcon, true);
            toolbar.insertBefore(separator, toolbar.firstChild);
          }
          return true;
        },

        showCalenderButtons: async function (showCal, showToDo, showToday) {
          let rw = Services.wm.getMostRecentWindow("mail:3pane");
          if (rw) {
            setCalenderButtons(rw, showCal, showToDo, showToday);
          }
          return true;
        },

        onMenuClick: new ExtensionCommon.EventManager({
          context,
          name: "msgHdrCustomizeApi.onMenuClick",
          register(fire) {
            function callback(event, id) {
              return fire.async(id);
            }

            msgHdrCustomizeWindowListener.add(callback);
            return function () {
              msgHdrCustomizeWindowListener.remove(callback);
            };
          }
        }).api(),

        onLoadMultiselect: new ExtensionCommon.EventManager({
          context,
          name: "msgHdrCustomizeApi.onLoadMultiselect",
          register(fire) {
            function callback(event, id) {
              return fire.async(id);
            }

            msgHdrCustomizeMultiselectListener.add(callback);
            return function () {
              msgHdrCustomizeMultiselectListener.remove(callback);
            };
          }
        }).api()
      }
    };
  }

  close() {
    for (let window of ExtensionSupport.openWindows) {
      if (window.location.href.indexOf("messenger.xhtml") != -1) {
        let toolbar = window.document.getElementById("header-view-toolbar");
        restoreDefaultStateForToolbar(toolbar, defaultOrderMSG);

        toolbar = window.document.getElementById("quick-filter-bar-collapsible-buttons");
        restoreDefaultStateForToolbar(toolbar, defaultOrderQF);

        setCalenderButtons(window, true, true, true);
      } else if (window.location.href.indexOf("messageWindow.xhtml") != -1) {
        let toolbar = window.document.getElementById("header-view-toolbar");
        restoreDefaultStateForToolbar(toolbar, defaultOrderMW);
      }
    }

    Services.obs.notifyObservers(null, "startupcache-invalidate", null);
  }
};

function updateDefaultOrder(toolbar, orgOrder, tmpIdPrefix) {
  let buttons = toolbar.childNodes;
  let newIdCnt = orgOrder.length - 1;
  for (let i = 0; i < buttons.length; i++) {
    let button = buttons[i];
    if (button.id) {
      if (orgOrder.indexOf(button.id)) {
        orgOrder.push(button.id);
      }
    } else {
      newIdCnt++;
      button.id = tmpIdPrefix + newIdCnt;
      orgOrder.push(button.id);
    }
  }
}

function setIconMode(obj, mode, margin) {
  let labels = obj.querySelectorAll(".toolbarbutton-text");
  let images = obj.querySelectorAll(".toolbarbutton-icon");
  let badgestack = obj.querySelectorAll(".toolbarbutton-badge-stack");

  for (let i = 0; i < labels.length; i++) {
    labels[i].collapsed = (mode == 1);
  }
  for (let i = 0; i < images.length; i++) {
    images[i].collapsed = (mode == 2);
  }
  for (let i = 0; i < badgestack.length; i++) {
    badgestack[i].collapsed = (mode == 2);
  }
}

function setIconMargin(obj, margin) {
  let labels = obj.querySelectorAll(".toolbarbutton-text");
  let images = obj.querySelectorAll(".toolbarbutton-icon");

  let val = margin ? margin + "px" : "";
  for (let i = 0; i < labels.length; i++) {
    labels[i].style.margin = val;
  }
  for (let i = 0; i < images.length; i++) {
    images[i].style.margin = val;
  }
}

function getButtonsInToolbar(rw, toolbar) {
  let ret = [];
  let buttons = toolbar ? toolbar.childNodes : [];
  for (let i = 0; i < buttons.length; i++) {
    let button = buttons[i];

    let compStyles = null;
    if (button.localName == "toolbaritem") { //for Smart Reply icon
      //let childButton = button.querySelectorAll(".msgHeaderView-button")[0];
      let childButton = button.querySelectorAll(".toolbarbutton-1")[0];
      compStyles = rw.getComputedStyle(childButton);
    } else {
      compStyles = rw.getComputedStyle(button);
    }

    if (button.id) {
      let iconURL = compStyles.getPropertyValue("list-style-image");
      let id = button.id.replace(/^col-/, "");
      let labels = button.querySelectorAll(".toolbarbutton-text");
      let images = button.querySelectorAll(".toolbarbutton-icon");

      let labelCol = labels[0] ? labels[0].collapsed : false;
      let imageCol = images[0] ? images[0].collapsed : false;
      let iconMode = 0;
      if (labelCol) {
        iconMode = 1;
      } else if (imageCol) {
        iconMode = 2;
      }

      let subLabels = [];
      //Smart Reply button etc has several labels
      for (let i = 0; i < labels.length; i++) {
        subLabels.push(labels[i].value);
      }

      let renamed = button.getAttribute("renamed") == "true";

      ret.push({
        id: id,
        label: button.getAttribute("label"),
        collapsed: button.collapsed,
        iconURL: iconURL,
        iconMode: iconMode,
        subLabels: subLabels,
        renamed: renamed
      });
    }
  }

  return ret;
}

function reorderButtonsMain(rw, order, toolbar, setIcon, spacerEnabled) {
  let buttons = toolbar.childNodes;
  let buttonObjs = [];
  for (let i = buttons.length - 1; i >= 0; i--) {
    let button = buttons[i];
    let id = button.id;
    if (id) {
      id = id.replace(/^col-/, "");
      button.id = id;
    } else {
      id = "tmpMhcId" + i;
    }

    buttonObjs[id] = button;
    toolbar.removeChild(button);
  }

  let reordered = [];
  for (let i = 0; i < order.length; i++) {
    let id = order[i].id;
    let obj = buttonObjs[id];
    if (obj) {
      obj.collapsed = order[i].collapsed;
      if (obj.id) { //tmpMhcId button does not have id in actual
        /* When Smart Reply button is collapsed, Reply button will not appear. This is because that TB thinks Smart Reply is there.
           So, change ID for a collapsed button.
           As a side effect of the ID changing, Quick Filter toolbar will not appear if you hide a button in the bar.
           For the reason, this is applied only for the "header-view-toolbar" only.
        */
        obj.id = (toolbar.id == "header-view-toolbar") && obj.collapsed ? "col-" + obj.id : obj.id;
      }

      setCustomLabel(obj, order[i]);

      //set icon display for each button
      if (setIcon && order[i].iconMode != undefined) {
        setIconMode(obj, order[i].iconMode);
      }

      toolbar.appendChild(obj);
    } else if (spacerEnabled && id.indexOf("mhcSPC") == 0) { //new Space or Separator
      obj = rw.document.createXULElement("toolbarspacer");
      obj.id = id;
      toolbar.appendChild(obj);
    } else if (spacerEnabled && id.indexOf("mhcSEP") == 0) {
      obj = rw.document.createXULElement("toolbarseparator");
      obj.id = id;
      toolbar.appendChild(obj);
    }
    reordered.push(id);
  }

  //add buttons not covered by order parameter or without ID
  for (let key in buttonObjs) {
    if (reordered.indexOf(key) == -1) {
      let obj = buttonObjs[key];
      if ((obj.id.indexOf("mhcSPC") != 0) && (obj.id.indexOf("mhcSEP") != 0)) {
        toolbar.appendChild(obj);
      }
    }
  }
}

function setCustomLabel(button, pref) {
  let labels = button.querySelectorAll(".toolbarbutton-text");
  let orgLabel = button.getAttribute("orgLabel");
  if (pref.renamed) {
    if (!orgLabel) { //first time rename
      button.setAttribute("orgLabel", button.getAttribute("label"));
    }

    for (let i = 0; i < labels.length; i++) {
      if (pref.subLabels[i]) {
        if (!orgLabel) {
          labels[i].setAttribute("orgLabel", labels[i].value);
        }
        labels[i].value = pref.subLabels[i];
      }
    }

    button.setAttribute("label", pref.label); //this shall not be performed the sub labels
    button.setAttribute("renamed", true);
  } else if (orgLabel) { //back to original name
    for (let i = 0; i < labels.length; i++) {
      let orgSubLabel = labels[i].getAttribute("orgLabel");
      if (orgSubLabel) {
        labels[i].value = orgSubLabel;
        labels[i].removeAttribute("orgLabel");
      }
    }

    button.setAttribute("label", orgLabel); //this shall not be performed the sub labels
    button.removeAttribute("orgLabel");
    button.removeAttribute("renamed");
  } //else keep original label
}

function restoreDefaultStateForToolbar(toolbar, orgOrder) {
  let buttons = toolbar.childNodes;
  let buttonObjs = {};
  let reorder = orgOrder.length > 0;
  for (let i = buttons.length - 1; i >= 0; i--) {
    let button = buttons[i];
    let id = button.id ? button.id.replace(/^col-/, "") : "rstrMhcId" + i;
    button.id = id;
    buttonObjs[id] = button;
    if (reorder) toolbar.removeChild(button);
  }

  for (let i = 0; i < orgOrder.length; i++) {
    let button = buttonObjs[orgOrder[i]];
    if (button && button.id.indexOf("mhc") != 0) {
      if (reorder) toolbar.appendChild(button);
      setCustomLabel(button, {
        renamed: false
      });
      delete buttonObjs[orgOrder[i]];
    }
  }

  for (let key in buttonObjs) {
    let button = buttonObjs[key];
    if (button && button.id.indexOf("mhc") != 0) { //excluding custom buttons and spacers
      if (reorder) toolbar.appendChild(button);
      setCustomLabel(button, {
        renamed: false
      });
    }
  }

  setIconMode(toolbar, 0);
  setIconMargin(toolbar, 0);
}

function setCalenderButtons(rw, showCal, showToDo, showToday) {
  let button = rw.document.getElementById("calendar-tab-button");
  button.collapsed = !showCal;

  button = rw.document.getElementById("task-tab-button");
  button.collapsed = !showToDo;
  button = rw.document.getElementById("calendar-status-todaypane-button");
  //button = rw.document.getElementById("calendar-show-todaypane-panel");
  button.collapsed = !showToday;
}

//Body function for button creation
function createNormalButton(rw, id, label, tooltip, observes, iconURL, command) {
  let button = rw.document.createXULElement("toolbarbutton");
  button.id = id;
  button.classList.add("toolbarbutton-1");
  button.classList.add("msgHeaderView-button");

  button.setAttribute("label", label);
  if (tooltip) {
    button.setAttribute("tooltiptext", tooltip);
  }

  if (observes) {
    button.setAttribute("observes", observes);
  }

  button.style.listStyleImage = 'url("' + iconURL + '")';
  button.addEventListener("command", command);
  button.collapsed = true; //hidden by default

  return button;
}

//Body function for menu-button creation
function createMenuButton(rw, id, label, labelForConfig, tooltip, observes, iconURL, command, popup) {
  let button = rw.document.createXULElement("toolbarbutton");
  button.id = id;
  button.setAttribute("is", "toolbarbutton-menu-button");
  button.setAttribute("type", "menu-button");

  button.setAttribute("label", labelForConfig); //not shown as button text. used for customize window.
  if (tooltip) {
    button.setAttribute("tooltiptext", tooltip);
  }

  button.classList.add("toolbarbutton-1");
  button.classList.add("msgHeaderView-button");
  button.style.listStyleImage = 'url("' + iconURL + '")';

  if (observes) {
    button.setAttribute("observes", observes);
  }

  if (command) {
    button.addEventListener("command", command);
  }
  button.collapsed = true; //hidden by default

  button.appendChild(popup);

  //this button is needed to reproduce menu-button
  let childButton = rw.document.createXULElement("toolbarbutton");
  childButton.classList.add("box-inherit");
  childButton.classList.add("toolbarbutton-menubutton-button");
  childButton.flex = 1;
  childButton.setAttribute("allowevents", true);
  childButton.setAttribute("label", label); //this label is shown as button text
  button.appendChild(childButton);

  let drop = rw.document.createXULElement("dropmarker");
  drop.setAttribute("type", "menu-button");
  drop.classList.add("toolbarbutton-menubutton-dropmarker");
  button.appendChild(drop);

  return button;
}

//Body function for popup-button, which always shows popup like the Tag button, creation.
function createPopupButton(rw, id, label, tooltip, observes, iconURL, popup) {
  let button = rw.document.createXULElement("toolbarbutton");
  button.id = id;
  button.classList.add("toolbarbutton-1");
  button.classList.add("msgHeaderView-button");
  button.setAttribute("type", "menu");
  button.setAttribute("label", label);
  button.setAttribute("tooltiptext", tooltip);
  button.style.listStyleImage = 'url("' + iconURL + '")';
  button.setAttribute("wantdropmarker", "true");
  button.collapsed = true; //hidden by default

  if (observes) {
    button.setAttribute("observes", observes);
  }

  button.appendChild(popup);

  return button;
}

function createConversationButton(rw, toolbar) {
  let button = createNormalButton(
    rw,
    "mhcConv",
    localeStrings["mhcConversation"],
    localeStrings["mhcConversationTooltip"],
    "button_showconversation",
    "chrome://messenger/skin/icons/conversation.svg",
    () => {
      rw.gConversationOpener.openConversationForMessages(rw.gFolderDisplay.selectedMessages);
      rw.RestoreFocusAfterHdrButton();
    }
  );

  toolbar.appendChild(button);
}

function createForwardAsButton(rw, toolbar) {
  let popup = rw.document.createXULElement("menupopup");
  popup.id = "mhcFwdAsPopup"
  let item = rw.document.createXULElement("menuitem");
  item.setAttribute("label", localeStrings["mhcForwardInline"]);
  item.setAttribute("tooltiptext", localeStrings["mhcForwardInlineTooltip"]);
  item.addEventListener("command", (event) => {
    rw.MsgForwardAsInline(event);
    event.stopPropagation();
    rw.RestoreFocusAfterHdrButton();
  });
  popup.appendChild(item);

  item = rw.document.createXULElement("menuitem");
  item.setAttribute("label", localeStrings["mhcForwardAttach"]);
  item.setAttribute("tooltiptext", localeStrings["mhcForwardAttachTooltip"]);
  item.addEventListener("command", (event) => {
    rw.MsgForwardAsAttachment(event);
    event.stopPropagation();
    rw.RestoreFocusAfterHdrButton();
  });
  popup.appendChild(item);

  let button = createMenuButton(
    rw,
    "mhcFwdAs",
    localeStrings["mhcForwardAs"],
    localeStrings["mhcForwardAs"],
    localeStrings["mhcForwardTooltip"],
    "button_forward",
    "chrome://messenger/skin/icons/forward.svg",
    (event) => {
      rw.MsgForwardMessage(event);
      rw.RestoreFocusAfterHdrButton();
    },
    popup
  );

  toolbar.appendChild(button);
}

function createPrintButton(rw, toolbar) {
  let popup = rw.document.createXULElement("menupopup");
  popup.id = "mhcPrintPopup"
  let item = rw.document.createXULElement("menuitem");
  item.setAttribute("label", localeStrings["mhcPrint"]);
  item.setAttribute("tooltiptext", localeStrings["mhcPrintTooltip"]);
  item.addEventListener("command", (event) => {
    rw.PrintEnginePrint();
    event.stopPropagation();
  });
  popup.appendChild(item);

  item = rw.document.createXULElement("menuitem");
  item.setAttribute("label", localeStrings["mhcPrintPreview"]);
  item.setAttribute("tooltiptext", localeStrings["mhcPrintPreviewTooltip"]);
  item.addEventListener("command", (event) => {
    rw.PrintEnginePrintPreview();
    event.stopPropagation();
  });
  popup.appendChild(item);
  
  let button = createMenuButton(
    rw,
    "mhcPrint",
    localeStrings["mhcPrint"],
    localeStrings["mhcPrint"],
    localeStrings["mhcPrintTooltip"],
    "button_print",
    "chrome://messenger/skin/icons/print.svg",
    () => {
      rw.PrintEnginePrint();
    },
    popup
  );

  toolbar.appendChild(button);
}

function createTagButton(rw, toolbar) {
  let popup = rw.document.createXULElement("menupopup");
  popup.addEventListener("popupshowing", () => {
    rw.InitMessageTags(popup);
  });
  popup.id = "mhcTagPopup";

  let menuitem = rw.document.createXULElement("menuitem");
  menuitem.id = "mhcTagPopup-addNewTag";
  menuitem.setAttribute("label", localeStrings["mhcNewTag"]);
  menuitem.setAttribute("command", "cmd_addTag");
  menuitem.setAttribute("accesskey", "N");
  popup.appendChild(menuitem);

  menuitem = rw.document.createXULElement("menuitem");
  menuitem.id = "mhcTagPopup-manageTags";
  menuitem.setAttribute("label", localeStrings["mhcManageTags"]);
  menuitem.setAttribute("command", "cmd_manageTags");
  menuitem.setAttribute("accesskey", "M");
  popup.appendChild(menuitem);

  menuitem = rw.document.createXULElement("menuseparator");
  popup.appendChild(menuitem);

  menuitem = rw.document.createXULElement("menuitem");
  menuitem.id = "mhcTagPopup-tagRemoveAll";
  menuitem.setAttribute("command", "cmd_removeTags");
  //label and accesskey for remove all are set by InitMessageTags(popup)
  popup.appendChild(menuitem);

  menuitem = rw.document.createXULElement("menuseparator");
  popup.appendChild(menuitem);

  let button  = createPopupButton(
    rw,
    "mhcTag",
    localeStrings["mhcTag"],
    localeStrings["mhcTagTooltip"],
    null,
    "chrome://messenger/skin/icons/tag.svg",
    popup
  );

  toolbar.appendChild(button);
}

function createStarButton(rw, toolbar) {
  let button = createNormalButton(
    rw,
    "mhcStar",
    localeStrings["mhcStar"],
    localeStrings["mhcStarTooltip"],
    null,
    "chrome://messenger/skin/icons/star.svg",
    () => {
      rw.MsgMarkAsFlagged();
    }
  );

  toolbar.appendChild(button);
}

function createPrevMsgButton(rw, toolbar) {
  let button = createNormalButton(
    rw,
    "mhcPrevMsg",
    "Previous",
    "Previous",
    //localeStrings["mhcStar"],
    //localeStrings["mhcStarTooltip"],
    null,
    "chrome://messenger/skin/icons/previousmsg.svg",
    () => {
      rw.goDoCommand('button_previousMsg');
    }
  );

  toolbar.appendChild(button);
}

var msgHdrCustomizeWindowListener = new class extends ExtensionCommon.EventEmitter {
  constructor() {
    super();
    this.callbackCount = 0;
  }

  handleEvent(event) {
    msgHdrCustomizeWindowListener.emit("menu-clicked", event.target.id);
  }

  add(callback) {
    this.on("menu-clicked", callback);
    this.callbackCount++;

    if (this.callbackCount == 1) {
      ExtensionSupport.registerWindowListener("msghdrcust_experimentListener", {
        chromeURLs: [
          "chrome://messenger/content/messenger.xhtml",
          "chrome://messenger/content/messageWindow.xhtml",
        ],
        onLoadWindow: function (window) {
          let isMessenger = window.location.href.indexOf("messenger.xhtml") != -1;
          let menuitemId = isMessenger ? "otherActionsCustomToolbarMSG" : "otherActionsCustomToolbarMW";
          let menuitem = window.document.getElementById(menuitemId);
          if (!menuitem) {
            let popup = window.document.getElementById("otherActionsPopup");
            let separator = window.document.createXULElement("menuseparator");
            separator.id = "mhcCustomizeSeparator";
            popup.appendChild(separator);
            menuitem = window.document.createXULElement("menuitem");
            menuitem.id = menuitemId;
            menuitem.label = localeStrings["mhcCustomizeToolbar"];
            menuitem.addEventListener("command", msgHdrCustomizeWindowListener.handleEvent);
            popup.appendChild(menuitem);
          }

          let popupset = window.document.getElementById("mainPopupSet");
          let menupopup = window.document.getElementById("mhcToolbarPopup");
          if (!menupopup) {
            menupopup = window.document.createXULElement("menupopup");
            menupopup.id = "mhcToolbarPopup";

            menuitem = window.document.createXULElement("menuitem");
            menuitem.id = isMessenger ? "mhcPopupCustomToolbarMSG" : "mhcPopupCustomToolbarMW";
            menuitem.label = localeStrings["mhcCustomizeToolbar"];
            menuitem.addEventListener("command", msgHdrCustomizeWindowListener.handleEvent);

            menupopup.appendChild(menuitem);
            popupset.appendChild(menupopup);

            let toolbox = window.document.getElementById("expandedHeadersTopBox");
            toolbox.setAttribute("context", "mhcToolbarPopup");
          }

          menupopup = window.document.getElementById("mhcToolbarPopupQF");
          if (!menupopup) {
            let qfToolbar = window.document.getElementById("quick-filter-bar");
            if (qfToolbar) {
              menupopup = window.document.createXULElement("menupopup");
              menupopup.id = "mhcToolbarPopupQF";

              menuitem = window.document.createXULElement("menuitem");
              menuitem.id = "mhcPopupCustomToolbarQF";
              menuitem.label = localeStrings["mhcCustomizeToolbar"];
              menuitem.addEventListener("command", msgHdrCustomizeWindowListener.handleEvent);

              menupopup.appendChild(menuitem);
              popupset.appendChild(menupopup);
              qfToolbar.setAttribute("context", "mhcToolbarPopupQF");
            }
          }

          menuitem = window.document.getElementById("mhcPopupCustomCAL");
          if (!menuitem) {
            menupopup = window.document.getElementById("toolbar-context-menu");
            menuitem = window.document.createXULElement("menuitem");
            menuitem.id = "mhcPopupCustomCAL";
            menuitem.label = localeStrings["mhcCustomizeCalendar"];
            menuitem.addEventListener("command", msgHdrCustomizeWindowListener.handleEvent);

            menupopup.appendChild(menuitem);
            popupset.appendChild(menupopup);
          }
        }
      });
    }
  }

  remove(callback) {
    this.off("menu-clicked", callback);
    this.callbackCount--;

    if (this.callbackCount == 0) {
      for (let window of ExtensionSupport.openWindows) {
        if ([
          "chrome://messenger/content/messenger.xhtml",
          "chrome://messenger/content/messageWindow.xhtml",
        ].includes(window.location.href)) {
          let menuitem = window.document.getElementById("otherActionsCustomToolbarMSG");
          if (!menuitem) menuitem = window.document.getElementById("otherActionsCustomToolbarMW");
          if (menuitem) {
            let popup = window.document.getElementById("otherActionsPopup");
            menuitem.removeEventListener("command", this.handleEvent);
            popup.removeChild(menuitem);
            let separator = window.document.getElementById("mhcCustomizeSeparator");
            popup.removeChild(separator);
          }

          let menupopup = window.document.getElementById("mhcToolbarPopup");
          let popupset = window.document.getElementById("mainPopupSet");
          if (menupopup) {
            let toolbox = window.document.getElementById("expandedHeadersTopBox");
            if (toolbox) toolbox.removeAttribute("context")
            popupset.removeChild(menupopup);
          }

          menupopup = window.document.getElementById("mhcToolbarPopupQF");
          if (menupopup) {
            let qfToolbar = window.document.getElementById("quick-filter-bar");
            if (qfToolbar) qfToolbar.removeAttribute("context")
            popupset.removeChild(menupopup);
          }

          menuitem = window.document.getElementById("mhcPopupCustomCAL");
          if (menuitem) {
            menupopup = window.document.getElementById("toolbar-context-menu");
            menupopup.removeChild(menuitem);
          }
        }
      }
      ExtensionSupport.unregisterWindowListener("msghdrcust_experimentListener");
    }
  }
};

var msgHdrCustomizeMultiselectListener = new class extends ExtensionCommon.EventEmitter {
  constructor() {
    super();
    this.callbackCount = 0;
  }

  handleEvent(event) {
    msgHdrCustomizeMultiselectListener.emit("multiselect", "multimessage");
  }

  add(callback) {
    this.on("multiselect", callback);
    this.callbackCount++;

    if (this.callbackCount == 1) {
      ExtensionSupport.registerWindowListener("msghdrcust_multiselectListener", {
        chromeURLs: [
          "chrome://messenger/content/messenger.xhtml"
        ],
        onLoadWindow: function (window) {
          let toolbox = window.document.getElementById("multimessage");
          toolbox.addEventListener("DOMContentLoaded", msgHdrCustomizeMultiselectListener.handleEvent);
        }
      });
    }
  }

  remove(callback) {
    this.off("multiselect", callback);
    this.callbackCount--;

    if (this.callbackCount == 0) {
      for (let window of ExtensionSupport.openWindows) {
        if ([
          "chrome://messenger/content/messenger.xhtml"
        ].includes(window.location.href)) {
          let toolbox = window.document.getElementById("multimessage");
          toolbox.removeEventListener("DOMContentLoaded", msgHdrCustomizeMultiselectListener.handleEvent);
        }
      }
      ExtensionSupport.unregisterWindowListener("msghdrcust_multiselectListener");
    }
  }
};
