var prefSuffix = "";
var reuseButtonList = false;
var listRebuildNeeded = false;

var iconModeDefault = {
  icon_mode: 1,
  icon_mode_mw: 1,
  icon_mode_qf: 0
};

async function init() {
  translate();

  browser.runtime.sendMessage({
    message: "GET_BUTTONS"
  }).then((message) => {
    initButtonList(message);
  });

  let select = document.mhc_form.icon_mode;
  select.addEventListener("change", () => {
    showIndividualOptions();
    setButtonTextAndStyle();
    updateIconModeButtons();
    updateIconAndButtonOrder(true, true);
  });

  select = document.mhc_form.button_size;
  select.addEventListener("change", () => {
    updateIconAndButtonOrder(true, true);
  });

  document.mhc_form.show_opt.value = "0";
  document.mhc_form.show_opt[0].addEventListener("change", () => {
    let changed = setIconModeForSelectedButtons(document.getElementById("placement"), 0);
    if (changed) {
      updateIconAndButtonOrder(false, true);
    }
  });
  document.mhc_form.show_opt[1].addEventListener("change", () => {
    let changed = setIconModeForSelectedButtons(document.getElementById("placement"), 1);
    if (changed) {
      updateIconAndButtonOrder(false, true);
    }
  });
  document.mhc_form.show_opt[2].addEventListener("change", () => {
    let changed = setIconModeForSelectedButtons(document.getElementById("placement"), 2);
    if (changed) {
      updateIconAndButtonOrder(false, true);
    }
  });

  let result = await browser.storage.local.get("apply_multi");
  let check = document.getElementById("apply_multi");
  check.checked = result.apply_multi ? true : false;
  check.addEventListener("change", () => {
    let checked = document.getElementById("apply_multi").checked;
    browser.storage.local.set({
      apply_multi: checked
    });
  });

  let buttonList = document.getElementById("placement");
  buttonList.addEventListener("change", () => {
    updateShowHideButton();
    updateIconModeButtons();
    enableButtons();
  });

  registerButtonCommands();

  document.getElementById("buttonClose").addEventListener("click", async () => {
    window.close();
  });
}

browser.runtime.onMessage.addListener((message) => {
  let aPromise = null;
  switch (message.message) {
    case "SEND_BUTTONS":
      aPromise = initButtonList(message);
      break;
    case "BUTTONS_UPDATED":
      if (listRebuildNeeded) {
        listRebuildNeeded = false;
        aPromise = initButtonList(message).then(result => {
          return true;
        });
      } else {
        aPromise = Promise.resolve(true);
      }
      break;
    default:
      aPromise = Promise.resolve(true);
      break;
  }

  return aPromise;
});

async function initButtonList(message) {
  if (!message.message) return;

  if (message.type == "normal") {
    prefSuffix = "";
  } else if (message.type == "messageDisplay") {
    prefSuffix = "_mw";
  } else if (message.type == "normal-QF") {
    prefSuffix = "_qf";
  }

  await setIconMode();

  let result = await browser.storage.local.get("icon_margin" + prefSuffix);
  select = document.mhc_form.button_size.value = result["icon_margin" + prefSuffix] ? result["icon_margin" + prefSuffix].toString() : "0";

  fillButtonList(message.buttons);
  setButtonTextAndStyle();
  if (prefSuffix == "") {
    document.getElementById("multiselArea").classList.remove("box-hidden");
    document.getElementById("multiselArea").classList.add("box-visible");
  }
  reuseButtonList = true;

  return true;
}

function setIconToOption(id, url) {
  let css = `
  #${id}:before{
    content: "";
    display: inline-block;
    width: 16px;
    height: 16px;
    background-image: ${url};
    background-repeat: no-repeat;
    background-size: contain;
    vertical-align: middle;
    padding-right: 5px;
  }`;

  document.styleSheets[0].insertRule(css);
}

async function setIconMode() {
  let result = await browser.storage.local.get("icon_mode" + prefSuffix);
  let select = document.mhc_form.icon_mode;
  select.value = result["icon_mode" + prefSuffix] === undefined ?
    iconModeDefault["icon_mode" + prefSuffix] :
    result["icon_mode" + prefSuffix];
  showIndividualOptions();
}

function fillButtonList(buttons) {
  let box = document.getElementById("placement");
  let preSelectedIndex = box.selectedIndex;

  for (let i = box.childNodes.length - 1; i >= 0; i--) {
    box.removeChild(box.childNodes[i]);
  }

  for (let i = 0; i < buttons.length; i++) {
    let button = buttons[i];
    let option = document.createElement("option");
    if (button.id.indexOf("mhcSPC") == 0) {
      option.textContent = browser.i18n.getMessage("mhcSpace");
    } else if (button.id.indexOf("mhcSEP") == 0) {
      option.textContent = browser.i18n.getMessage("mhcSeparator");
    } else if (!reuseButtonList) {
      setIconToOption(button.id, button.iconURL);
    }
    option.id = button.id;
    option.setAttribute("buttonLabel", button.label);
    option.setAttribute("buttonCollapsed", button.collapsed);
    option.setAttribute("buttonIconMode", button.iconMode);
    option.setAttribute("buttonId", button.id);
    option.setAttribute("buttonSubLabels", button.subLabels.join(", "));
    option.setAttribute("buttonRenamed", button.renamed);

    if (button.collapsed) {
      option.classList.add("option-hidden");
    }

    box.appendChild(option);
  }

  box.selectedIndex = preSelectedIndex;
  
  enableButtons();
}

function registerButtonCommands() {
  document.getElementById("buttonUp").addEventListener("click", () => {
    let moved = moveOptsInSelection(
      document.getElementById("placement"),
      true
    );
    if (moved) updateIconAndButtonOrder(false, true);
  });

  document.getElementById("buttonDown").addEventListener("click", () => {
    let moved = moveOptsInSelection(
      document.getElementById("placement"),
      false
    );
    if (moved) updateIconAndButtonOrder(false, true);
  });

  document.getElementById("buttonHide").addEventListener("click", () => {
    let changed = hideSelectedButtons(
      document.getElementById("placement")
    );
    if (changed) {
      updateShowHideButton();
      updateIconAndButtonOrder(false, true);
    }
  });

  document.getElementById("buttonSpc").addEventListener("click", () => {
    let changed = addSpacer(
      document.getElementById("placement"),
      "mhcSPC"
    );
    if (changed) updateIconAndButtonOrder(false, true);
  });

  document.getElementById("buttonSep").addEventListener("click", () => {
    let changed = addSpacer(
      document.getElementById("placement"),
      "mhcSEP"
    );
    if (changed) updateIconAndButtonOrder(false, true);
  });

  document.getElementById("buttonReset").addEventListener("click", () => {
    resetButtonOrder();
  });

  document.getElementById("buttonRename").addEventListener("click", () => {
    let changed = renameButtons(document.getElementById("placement"));
    if (changed) {
      /* rebuild required for the case getting back original name */
      listRebuildNeeded = true;
      updateIconAndButtonOrder(false, true);
    }
  });
}

function showIndividualOptions() {
  let mode = parseInt(document.mhc_form.icon_mode.value);
  let box = document.getElementById("customShowOptArea");
  if (mode == -1) {
    box.classList.remove("box-collapse");
    box.classList.add("box-visible");
  } else {
    box.classList.add("box-collapse");
    box.classList.remove("box-visible");
  }
}

function setButtonTextAndStyle() {
  let hiddenMsg = browser.i18n.getMessage("mhcHidden");
  let mode = parseInt(document.mhc_form.icon_mode.value)

  let select = document.getElementById("placement");

  for (let i = 0; i < select.options.length; i++) {
    let option = select.options[i];
    if (option.id.indexOf("mhcSPC") == 0 || option.id.indexOf("mhcSEP") == 0) {
      continue;
    }

    let label = option.getAttribute("buttonLabel");
    if (mode >= 0) {
      option.setAttribute("buttonIconMode", mode);
    }
    if (option.getAttribute("buttonCollapsed") == "true") {
      option.textContent = label + " " + hiddenMsg;
      option.id = option.getAttribute("buttonId");
      option.id = "hide-" + option.getAttribute("buttonId");
    } else if (option.getAttribute("buttonIconMode") == "0") {
      option.textContent = label;
      option.id = option.getAttribute("buttonId");
    } else if (option.getAttribute("buttonIconMode") == "1") {
      option.textContent = "(" + label + ")";
      option.id = option.getAttribute("buttonId");
    } else if (option.getAttribute("buttonIconMode") == "2") {
      option.textContent = label;
      option.id = "noicon-" + option.getAttribute("buttonId");
    }
  }
}

function enableButtons() {
  let select = document.getElementById("placement");
  let opts = select.selectedOptions;

  if (opts.length > 0) {
    document.getElementById("buttonHide").disabled = false;
    document.getElementById("buttonSpc").disabled = (opts.length > 1);
    document.getElementById("buttonSep").disabled = (opts.length > 1);
    let firstIndex = getIndexOfOption(select, opts[0]);
    document.getElementById("buttonUp").disabled = (firstIndex == 0);
    let lastIndex = getIndexOfOption(select, opts[opts.length - 1]);
    document.getElementById("buttonDown").disabled = (lastIndex == select.options.length - 1);
    document.getElementById("buttonRename").disabled = false;
  } else {
    document.getElementById("buttonUp").disabled = true;
    document.getElementById("buttonDown").disabled = true;
    document.getElementById("buttonHide").disabled = true;
    document.getElementById("buttonSpc").disabled = true;
    document.getElementById("buttonSep").disabled = true;
    document.getElementById("buttonRename").disabled = true;
    document.mhc_form.show_opt[0].disabled = true;
    document.mhc_form.show_opt[1].disabled = true;
    document.mhc_form.show_opt[2].disabled = true;
  }
}

function moveOptsInSelection(list, up) {
  if (list.selectedIndex == -1) return false;
  let orgOpts = list.selectedOptions;

  //copy to array. selectedOptions will change by the appendChild() and insertBefore()
  let opts = [];
  for (let i = 0; i < orgOpts.length; i++) {
    opts.push(orgOpts[i]);
  }

  let firstIndex = getIndexOfOption(list, opts[0]);
  let lastIndex = getIndexOfOption(list, opts[opts.length - 1]);
  if (up && firstIndex == 0) return false;
  if (!up && lastIndex == list.childNodes.length - 1) return false;

  if (up) {
    for (let i = 0; i < opts.length; i++) {
      let opt = opts[i];
      swapOptInSelection(list, up, opt);
    }
  } else {
    for (var i = opts.length - 1; i >= 0; i--) {
      let opt = opts[i];
      swapOptInSelection(list, up, opt);
    }
  }

  enableButtons();

  return true;
}

function swapOptInSelection(list, up, opt) {
  let refIndex = up ? getIndexOfOption(list, opt) - 1 : getIndexOfOption(list, opt) + 2;
  let refChild = list.childNodes[refIndex];
  if (refIndex >= list.childNodes.length) {
    list.appendChild(opt);
  } else {
    list.insertBefore(opt, refChild);
  }
}

function getIndexOfOption(list, opt) {
  let index = -1;
  for (let i = 0; i < list.length; i++) {
    if (list[i] === opt) {
      index = i;
      break;
    }
  }

  return index;
}

//save prefs to local storage
function updateIconAndButtonOrder(icon, button) {
  let obj = {};
  
  if (icon) {
    let mode = document.mhc_form.icon_mode.value;
    obj["icon_mode" + prefSuffix] = parseInt(mode);
    let margin = document.mhc_form.button_size.value;
    obj["icon_margin" + prefSuffix] = parseInt(margin);
  }
  
  if (button) {
    let box = document.getElementById("placement");
    let newOrder = [];
    for (let i = 0; i < box.options.length; i++) {
      let opt = box.options[i];
      let subLabels = [];
      if (opt.getAttribute("buttonSubLabels")) {
        subLabels = opt.getAttribute("buttonSubLabels").split(/\s*,\s*/);
      }

      newOrder.push({
        id: opt.getAttribute("buttonId"),
        label: opt.getAttribute("buttonLabel"),
        collapsed: (opt.getAttribute("buttonCollapsed") == "true"),
        iconMode: parseInt(opt.getAttribute("buttonIconMode")),
        subLabels: subLabels,
        renamed: (opt.getAttribute("buttonRenamed") == "true"),
      });
    }

    obj["placement" + prefSuffix] = newOrder;
  }
  
  if (icon || button) browser.storage.local.set(obj);
}

function hideSelectedButtons(list) {
  if (list.selectedIndex == -1) return false;

  let hiddenMsg = browser.i18n.getMessage("mhcHidden");
  let opts = list.selectedOptions;

  let hide = !(opts[0].getAttribute("buttonCollapsed") == "true"); //adopt top button for hidden mode
  for (let i = opts.length - 1; i >= 0; i--) {
    let opt = opts[i];
    if (hide) {
      if (opt.id.indexOf("mhcSPC") == 0 || opt.id.indexOf("mhcSEP") == 0) {
        list.removeChild(opt);
      } else {
        opt.textContent = opt.getAttribute("buttonLabel") + " " + hiddenMsg;
        opt.setAttribute("buttonCollapsed", true);
        opt.classList.add("option-hidden");
        setButtonTextAndStyle();
        updateIconModeButtons();
      }
    } else {
      opt.textContent = opt.getAttribute("buttonLabel");
      opt.setAttribute("buttonCollapsed", false);
      opt.classList.remove("option-hidden");
      setButtonTextAndStyle();
      updateIconModeButtons();
    }
  }

  return true;
}

function renameButtons(list) {
  if (list.selectedOptions.length != 1) return false;
  let opt = list.selectedOptions[0];
  let label = opt.getAttribute("buttonSubLabels") ? opt.getAttribute("buttonSubLabels") : opt.getAttribute("buttonLabel");
  let numLabels = label.split(/\s*,\s*/).length;
  let msg = browser.i18n.getMessage("mhcRenameMsg");
  let newLabel = window.prompt(msg, label);
  let renamed = false;
  if (newLabel && newLabel.split(/\s*,\s*/).length == numLabels) {
    let newLabelArray = newLabel.split(/\s*,\s*/);
    opt.setAttribute("buttonSubLabels", newLabel);
    opt.setAttribute("buttonLabel", newLabelArray[0]);
    opt.setAttribute("buttonRenamed", true);
    renamed = true;
  } else if (newLabel == "") { //disable rename
    opt.setAttribute("buttonRenamed", false);
    opt.removeAttribute("buttonSubLabels");
    renamed = true;
  }

  return renamed;
}

function setIconModeForSelectedButtons(list, mode) {
  if (list.selectedIndex == -1) return false;
  let opts = list.selectedOptions;
  for (let i = 0; i < opts.length; i++) {
    let opt = opts[i];
    opt.setAttribute("buttonIconMode", mode);
  }

  setButtonTextAndStyle();

  return true;
}

async function resetButtonOrder() {
  let obj = {};

  obj["icon_mode" + prefSuffix] = 1;
  obj["icon_margin" + prefSuffix] = 0;
  obj["placement" + prefSuffix] = [];
  if (!prefSuffix) obj["apply_multi"] = false;

  await browser.storage.local.set(obj);

  let msg = browser.i18n.getMessage("mhcRestartMsg");
  alert(msg);
}

function addSpacer(list, type) {
  if (list.selectedIndex == -1) return false;
  let refOpt = list.selectedOptions[0];
  let label = "";
  if (type == "mhcSPC") {
    label = browser.i18n.getMessage("mhcSpace");
  } else if (type == "mhcSEP") {
    label = browser.i18n.getMessage("mhcSeparator");
  }
  let opt = document.createElement("option");
  opt.textContent = label;
  opt.setAttribute("buttonCollapsed", false);
  let date = new Date();
  let id = date.getTime();
  opt.id = type + "-" + id;
  opt.setAttribute("buttonLabel", label);
  opt.setAttribute("buttonId", opt.id);
  opt.setAttribute("buttonIconMode", 0);

  list.insertBefore(opt, refOpt);

  return true;
}

function updateShowHideButton() {
  let box = document.getElementById("placement");
  let opt = box.selectedOptions[0];
  let localeKey = "";
  if (opt) {
    let spacerSelected = opt.id.indexOf("mhcSPC") == 0 || opt.id.indexOf("mhcSEP") == 0;
    if (spacerSelected) {
      localeKey = "mhcDel";
    } else {
      let hide = !(opt.getAttribute("buttonCollapsed") == "true");
      localeKey = hide ? "mhcHide" : "mhcShow";
    }
  } else { //hiding space and separator case. opt is removed.
    localeKey = "mhcHide";
  }

  document.getElementById("buttonHide").value = browser.i18n.getMessage(localeKey);
}

function updateIconModeButtons() {
  let box = document.getElementById("placement");
  let opt = box.selectedOptions[0];
  if (!opt) return false;

  let mode = 0;
  let modeButtonDisabled = opt.id.indexOf("mhcSPC") == 0 || opt.id.indexOf("mhcSEP") == 0;
  if (modeButtonDisabled || opt.getAttribute("buttonCollapsed") == "true") {
    mode = -1;
  } else {
    mode = parseInt(opt.getAttribute("buttonIconMode"));
  }

  switch (mode) {
    case -1:
      document.mhc_form.show_opt.value = "0";
      document.mhc_form.show_opt[0].disabled = true;
      document.mhc_form.show_opt[1].disabled = true;
      document.mhc_form.show_opt[2].disabled = true;
      break;
    case 0:
      document.mhc_form.show_opt.value = "0";
      document.mhc_form.show_opt[0].disabled = false;
      document.mhc_form.show_opt[1].disabled = false;
      document.mhc_form.show_opt[2].disabled = false;
      break;
    case 1:
      document.mhc_form.show_opt.value = "1";
      document.mhc_form.show_opt[0].disabled = false;
      document.mhc_form.show_opt[1].disabled = false;
      document.mhc_form.show_opt[2].disabled = false;
      break;
    case 2:
      document.mhc_form.show_opt.value = "2";
      document.mhc_form.show_opt[0].disabled = false;
      document.mhc_form.show_opt[1].disabled = false;
      document.mhc_form.show_opt[2].disabled = false;
      break;
    default:
      document.mhc_form.show_opt.value = "0";
      document.mhc_form.show_opt[0].disabled = true;
      document.mhc_form.show_opt[1].disabled = true;
      document.mhc_form.show_opt[2].disabled = true;
      break;
  }
}

init();
